# -*- coding: utf-8 -*-
#
# Synchronize Groups to OPSI
#
# Copyright 2013-2014 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

name = 'opsigrouplistener'
description = 'listener to add, remove and update groups with windows clients in OPSI'
filter = '(objectClass=univentionGroup)'
attributes = ['uniqueMember']

__package__ = ''  # workaround for PEP 366

import simplejson as json  # the "normal" json module doesn't work in optimize mode, no idea why
import traceback
import listener
from univention.debug import debug, LISTENER, INFO, ERROR, WARN
import univention.uldap

syncGroupsContainingObjectsMatchingFilter = listener.baseConfig.get('opsi/grouplistener/host/filter', '(objectClass=univentionWindows)')

from OPSI.Backend.BackendManager import *
import inspect

listener.setuid(0)
backend = BackendManager(
		dispatchConfigFile=u'/etc/opsi/backendManager/dispatch.conf',
		backendConfigDir=u'/etc/opsi/backends',
		extensionConfigDir=u'/etc/opsi/backendManager/extend.d',
		depotBackend=True,
		hostControlBackend=True,
		hostControlSafeBackend=True
		)
listener.unsetuid()
backend_methods = dict(inspect.getmembers(backend, callable))


ldap_info = {}
def opsi_admin(*args, **kwargs):
	""" Run opsi-admin -d method with any number of arguments.

	:args: All arguments are passed to opsi-admin -d method
	:returns: Return opsi-admin process object
	"""
	assert args

	method = backend_methods.get(args[0], None)
	if not method:
		raise ValueError('Method not found.')
	listener.setuid(0)
	res = method(*args[1:], **kwargs)
	listener.unsetuid()
	return res


def get_groupstobject(groupid, ignore_errors=False):
	"""Retrieve JSON data representing hostid

	:hostid: FQDN of the requested host
	:returns: None or JSON data representing hostid
	"""

	try:
		# query opsi for existing information about client
		res = [ json.loads(host.toJson()) for host in opsi_admin('group_getObjects', [], id=groupid) ]
		if res:
			return res[0]
	except Exception:
		if not ignore_errors:
			debug(LISTENER, ERROR, traceback.format_exc())


def ldap_reconnect():
	debug(LISTENER, INFO, '%s: ldap reconnect triggered' % name)
	if 'ldapserver' in ldap_info and 'basedn' in ldap_info and 'binddn' in ldap_info and 'bindpw' in ldap_info:
		try:
			ldap_info['lo'] = univention.uldap.access(host = ldap_info['ldapserver'],
					base = ldap_info['basedn'],
					binddn = ldap_info['binddn'],
					bindpw = ldap_info['bindpw'],
					start_tls = 2)
		except Exception, ex:
			debug(LISTENER, ERROR, '%s: ldap reconnect failed: %s' % (name,str(ex)))
			ldap_info['lo'] = None
		else:
			if ldap_info['lo'] == None:
				debug(LISTENER, ERROR, '%s: ldap reconnect failed' % name)


def setdata(key, value):
	if key == 'bindpw':
		debug(LISTENER, INFO, '%s: listener passed key="%s" value="<HIDDEN>"' % (name, key))
	else:
		debug(LISTENER, INFO, '%s: listener passed key="%s" value="%s"' % (name, key, value))

	if key in [ 'ldapserver', 'basedn', 'binddn', 'bindpw' ]:
		ldap_info[ key ] = value
	else:
		debug(LISTENER, INFO, '%s: listener passed unknown data (key="%s" value="%s")' % (name, key, value))

	if key == 'ldapserver':
		debug(LISTENER, INFO, '%s: ldap server changed to %s' % (name, value))
		ldap_reconnect()


def initialize():
	debug(LISTENER, INFO, '%s: Initialize' % name)
	if not ldap_info['lo']:
		ldap_reconnect()


def get_cn(dn):
	""" return the hostname of a computer object
	"""
	return univention.uldap.explodeDn(dn, True)[0]


def rename_group(new, old):
	old_groupid = old['cn'][0].lower()
	debug(LISTENER, WARN, '%s: renaming group: %s' % (name, old_groupid))
	# remove old group
	if get_groupstobject(old_groupid):
		opsi_admin('group_delete', old_groupid)
	else:
		debug(LISTENER, ERROR, '%s: renaming group, group not found: %s' % (name, old_groupid))
	# create new group
	ensure_group_exists(new)
	# add computers to group
	if 'uniqueMember' in new:
		computers = match_filter_and_return_cns(new['uniqueMember'])
		add_computers_to_group(computers, new['cn'][0].lower())


def match_filter_and_return_cns(dn_list):
	# ensure the objects match filter syncGroupsContainingObjectsMatchingFilter(default (objectClass=univentionWindows))
	if not ldap_info['lo']:
		ldap_reconnect()
	if not len(dn_list) > 0:
		return None
	searchfilter = "(&%s(|" % syncGroupsContainingObjectsMatchingFilter
	for dn in dn_list:
		if not len(dn) < 1:
			if dn.startswith("uid"):
				continue
			else:
				searchfilter += "(cn=%s)" % get_cn(dn)
	searchfilter += "))"
	debug(LISTENER, INFO, '%s: searchfilter: %s' % (name, searchfilter))

	result = ldap_info['lo'].search(filter=searchfilter, attr=['cn'])
	computerobject_cn_list = []
	if result:
		debug(LISTENER, INFO, '%s: result: %s' % (name, result))
	for computerobject in result:
		computerobject_cn_list.append(computerobject[1]['cn'][0])
	debug(LISTENER, INFO, '%s: objects that will be added/removed: %s' % (name, computerobject_cn_list))
	return computerobject_cn_list


def ensure_group_exists(groupobject):
	groupid = groupobject['cn'][0].lower()
	debug(LISTENER, INFO, '%s: creating group: %s' % (name, groupid))
	# creating an existing group does nothing if it already exists, so in order to ensure existance, always create the group
	desc = ""
	if 'description' in groupobject:
		desc = groupobject['description']
	if not get_groupstobject(groupid):
		opsi_admin('group_createHostGroup', groupid, desc)
	else:
		debug(LISTENER, ERROR, '%s: creating group, group already exists: %s' % (name, groupid))


def add_computers_to_group(cn_list, groupid):
	# Add computer names to opsi group
	debug(LISTENER, INFO, '%s: adding computers: %s' % (name, cn_list))
	domainname = listener.baseConfig.get('domainname', '').lower()
	if get_groupstobject(groupid):
		for cn in cn_list:
			opsi_admin('objectToGroup_create', 'HostGroup', groupid, '%s.%s' % (cn.lower(), domainname))
	else:
		debug(LISTENER, ERROR, '%s: adding computers, group not found: %s' % (name, groupid))


def remove_computers_from_group(cn_list, groupid):
	# Remove computer names from opsi group
	debug(LISTENER, INFO, '%s: removing computers: %s' % (name, cn_list))
	domainname = listener.baseConfig.get('domainname', '').lower()
	if get_groupstobject(groupid):
		for cn in cn_list:
			opsi_admin('objectToGroup_delete', 'HostGroup', groupid, '%s.%s' % (cn.lower(), domainname))
	else:
		debug(LISTENER, ERROR, '%s: removing computers, group not found: %s' % (name, groupid))


# does not support nested groups
def handler(dn, new, old):
	ldap_reconnect()
	debug(LISTENER, INFO, '%s: newgrp: %s\noldgrp: %s' % (name, new, old))
	# remove group if old and not new
	if old and not new:
		old_groupid = old['cn'][0].lower()
		debug(LISTENER, INFO, '%s: deleting group: %s' % (name, old_groupid))
		if get_groupstobject(old_groupid):
			opsi_admin('group_delete', old_groupid)
		else:
			debug(LISTENER, ERROR, '%s: deleting group, group not found: %s' % (name, old_groupid))
		return

	# Group rename
	new_groupid = new['cn'][0].lower()
	if 'cn' in new and 'cn' in old:
		if new['cn'][0] != old['cn'][0]:
			rename_group(new, old)

	newmember = []
	if 'uniqueMember' in new:
		newmember = match_filter_and_return_cns(new['uniqueMember'])
		debug(LISTENER, INFO, '%s: newmember %s' % (name, newmember))

	oldmember = []
	if 'uniqueMember' in old:
		oldmember = match_filter_and_return_cns(old['uniqueMember'])
		debug(LISTENER, INFO, '%s: oldmember: %s' % (name, oldmember))

	# has the group been modified?
	diffdel = list(set(oldmember) - set(newmember))
	diffadd = list(set(newmember) - set(oldmember))

	if len(diffadd) > 0:
		if len(diffadd) > 0:
			ensure_group_exists(new)
			add_computers_to_group(diffadd, new_groupid)
	if len(diffdel) > 0:
		if len(diffdel) > 0:
			remove_computers_from_group(diffdel, old['cn'][0].lower())

	# delete empty groups
	if not newmember:
		debug(LISTENER, INFO, '%s: deleting empty group: %s' % (name, new_groupid))
		if get_groupstobject(new_groupid):
			opsi_admin('group_delete', new_groupid)
		else:
			debug(LISTENER, ERROR, '%s: deleting empty group, group not found: %s' % (name, new_groupid))

# vi: ft=python:tw=79:sw=4:ts=4:noet
