# -*- coding: utf-8 -*-
#
# Register Windows Systems at OPSI
#
# Copyright 2012 DIGITEC GmbH based on work on Univention Copyright 2012-2014 Univention GmbH
#
# http://www.digitec.de/
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

name = 'opsilistener'
description = 'listener to add, remove and update windows clients in OPSI'
filter = '(objectClass=univentionWindows)'
attributes = ['macAddress', 'aRecord', 'cn', 'univentionInventoryNumber', 'description']
modrdn = '1'

RENAME = False

__package__ = ''  # workaround for PEP 366

import simplejson as json  # the "normal" json module doesn't work in optimize mode, no idea why
import traceback
import listener
from univention.debug import debug, LISTENER, INFO, ERROR, WARN
import univention.uldap

filter = listener.baseConfig.get('digitec/opsi/listener/host/filter', filter)

from OPSI.Backend.BackendManager import *
import inspect

listener.setuid(0)
backend = BackendManager(
		dispatchConfigFile=u'/etc/opsi/backendManager/dispatch.conf',
		backendConfigDir=u'/etc/opsi/backends',
		extensionConfigDir=u'/etc/opsi/backendManager/extend.d',
		depotBackend=True,
		hostControlBackend=True,
		hostControlSafeBackend=True
		)
listener.unsetuid()
backend_methods = dict(inspect.getmembers(backend, callable))


ldap_info = {}


def opsi_admin(*args, **kwargs):
	""" Run opsi-admin -d method with any number of arguments.

	:args: All arguments are passed to opsi-admin -d method
	:returns: Return opsi-admin process object
	"""
	assert args

	method = backend_methods.get(args[0], None)
	if not method:
		raise ValueError('Method not found.')
	listener.setuid(0)
	res = method(*args[1:], **kwargs)
	listener.unsetuid()
	return res


def get_hostobject(hostid, ignore_errors=False):
	"""Retrieve JSON data representing hostid

	:hostid: FQDN of the requested host
	:returns: None or JSON data representing hostid
	"""

	try:
		# query opsi for existing information about client
		res = [json.loads(host.toJson()) for host in opsi_admin('host_getObjects', [], id=hostid)]
		if res:
			return res[0]
	except Exception:
		if not ignore_errors:
			debug(LISTENER, ERROR, traceback.format_exc())


def get_all_hostobjects():
	"""Retrieve  JSON data representing all hosts

	:returns: None or JSON data representing hosts
	"""

	# query opsi for existing information about client
	try:
		return [json.loads(host.toJson()) for host in opsi_admin('host_getObjects', [], type="OpsiClient")]
	except Exception:
		debug(LISTENER, ERROR, traceback.format_exc())
	return []


def update_host(hostid, new, old, host_obj=None):
	"""Update an OPSI host object

	:hostid: FQDN of the requested OPSI host
	:new: dictionary as it is returned by LDAP search
	:old: dictionary as it is returned by LDAP search
	:host_obj: optional host object as it is returned by get_hostobject
	:returns: returns result of opsi_admin call
	"""

	debug(LISTENER, INFO, '%s: updating OPSI client: %s' % (name, hostid))

	mapping_ldap_opsi = {
		'aRecord':                    'ipAddress',
		'macAddress':                 'hardwareAddress',
		'description':                'description',
		'univentionInventoryNumber':  'inventoryNumber'}
	force_create = False

	if not host_obj:
		host_obj = get_hostobject(hostid.lower())

	if not host_obj:
		debug(LISTENER, WARN, '%s: OPSI client not found, creating it: %s' %
				(name, hostid))
		host_obj = {'id': hostid}
		force_create = True

	changed = 0
	for i in ('aRecord', 'macAddress', 'description',
			'univentionInventoryNumber'):
		attr_new = new.get(i, [None])[0]
		attr_old = old.get(i, [None])[0]
		if attr_new != attr_old or force_create:
			debug(LISTENER, INFO, '%s: updating OPSI attribute %s for host %s to %s' %
					(name, mapping_ldap_opsi[i], hostid, attr_new))
			if attr_new:
				# encode string to allow OPSI to decode it properly
				attr_new = attr_new.encode()
			host_obj[mapping_ldap_opsi[i]] = attr_new
			changed = 1
		else:
			# update all attributes from LDAP to create a sync state
			if attr_new:
				# encode string to allow OPSI to decode it properly
				attr_new = attr_new.encode()
			host_obj[mapping_ldap_opsi[i]] = attr_new

	if force_create:
		create_host(host_obj['id'], host_obj['description'],
				host_obj['hardwareAddress'], host_obj['ipAddress'],
				host_obj['inventoryNumber'])
	elif changed:
		# WORKAROUND: host_updateObjects doesn't work when setting
		# attributes like ipAddress and hardwareAddress to None, therefore
		# use host_createObjects. By pulling the object first via
		# opsi-admin and updating it on the way, last seen and created
		# timestamps are kept
		if listener.baseConfig.is_true('digitec/opsi/listener/host/modify', True):
			debug(LISTENER, INFO, '%s: modifying OPSI client: %s' % (name, hostid))
			opsi_admin("host_createObjects", json.dumps(host_obj))
		else:
			debug(LISTENER, INFO, '%s: modify OPSI client: %s skipped' % (name, hostid))
	else:
		debug(LISTENER, INFO, '%s: no changes found for OPSI client: %s' % (name, hostid))


def delete_host(hostid):
	if listener.baseConfig.is_true('digitec/opsi/listener/host/delete', True):
		debug(LISTENER, INFO, '%s: deleting OPSI client: %s' % (name, hostid))
		opsi_admin("host_delete", hostid)
	else:
		debug(LISTENER, INFO, '%s: deleting of OPSI client skipped: %s' % (name, hostid))


def create_host(hostid, desc, mac, ip, inventory_nr):
	if listener.baseConfig.is_true('digitec/opsi/listener/host/create', True):
		debug(LISTENER, INFO, '%s: creating OPSI client: %s' % (name, hostid))
		opsi_admin("host_createOpsiClient", hostid,
				description=desc, hardwareAddress=mac, ipAddress=ip, inventoryNumber=inventory_nr)
	else:
		debug(LISTENER, INFO, '%s: creating of OPSI client skipped: %s' % (name, hostid))


def ldap_reconnect():
	debug(LISTENER, INFO, '%s: ldap reconnect triggered' % name)
	if 'ldapserver' in ldap_info and 'basedn' in ldap_info and 'binddn' in ldap_info and 'bindpw' in ldap_info:
		try:
			ldap_info['lo'] = univention.uldap.access(host=ldap_info['ldapserver'],
					base=ldap_info['basedn'],
					binddn=ldap_info['binddn'],
					bindpw=ldap_info['bindpw'],
					start_tls=2)
		except Exception, ex:
				debug(LISTENER, ERROR, '%s: ldap reconnect failed: %s' % (name, str(ex)))
				ldap_info['lo'] = None
	else:
			if ldap_info['lo'] is None:
					debug(LISTENER, ERROR, '%s: ldap reconnect failed' % name)


def setdata(key, value):
	if key == 'bindpw':
		debug(LISTENER, INFO, '%s: listener passed key="%s" value="<HIDDEN>"' % (name, key))
	else:
		debug(LISTENER, INFO, '%s: listener passed key="%s" value="%s"' % (name, key, value))

	if key in ['ldapserver', 'basedn', 'binddn', 'bindpw']:
		ldap_info[key] = value
	else:
		debug(LISTENER, INFO, '%s: listener passed unknown data (key="%s" value="%s")' % (name, key, value))

	if key == 'ldapserver':
		debug(LISTENER, INFO, '%s: ldap server changed to %s' % (name, value))
		ldap_reconnect()


def initialize():
	debug(LISTENER, INFO, '%s: Initialize' % name)
	if not ldap_info['lo']:
		ldap_reconnect()
	if ldap_info['lo']:
		for host in get_all_hostobjects():
			hostid = host['id'].lower()
			debug(LISTENER, INFO, '%s: Found OPSI-Host: %s' % (name, hostid))
			domainname = listener.baseConfig.get('domainname', '')
			cn = hostid.replace(".%s" % domainname, "")
			search = '(&%s(cn=%s))' % (filter, cn)
			debug(LISTENER, INFO, '%s: Search OPSI-Host: %s' % (name, search))
			if not ldap_info['lo'].search(filter=search):
				delete_host(hostid)


def handler(dn, new, old, command):
	global RENAME
	domainname = listener.baseConfig.get('domainname', '')
	# object modified
	if command == 'm':
		if new and old:
			cn = new['cn'][0]
			update_host(('.'.join((cn, domainname)).lower()), new, old)
		else:
			debug(LISTENER, ERROR, '%s: Modify not possible, (old and new == False). new: %r, old: %r' % (name, new, old))

	# object deleted
	elif command == 'd':
		if old:
			cn = old['cn'][0]
			hostid = '.'.join((cn, domainname)).lower()

			delete_host(hostid)
		else:
			debug(LISTENER, ERROR, '%s: Delete not possible, (old == False). old: %r' % (name, old))

	# object created
	elif command == 'a':
		if RENAME:
			debug(LISTENER, INFO, '%s: Detected rename of client, destination DN: %s' % (name, dn))
			RENAME = False
		elif new:
			cn = new['cn'][0]
			hostid = '.'.join((cn, domainname)).lower()
			host_obj = get_hostobject(hostid, ignore_errors=True)

			if host_obj:
				# update client instead of creating a new one, this will prevent
				# the creation of a new OPSI host key
				update_host(hostid, new, {}, host_obj=host_obj)

			else:
				# create a new client, this will also the creation of a new OPSI
				# host key, even if the object already exists
				ip = new.get('aRecord', [None])[0]
				mac = new.get('macAddress', [None])[0]
				desc = new.get('description', [None])[0]
				inventory_nr = new.get('univentionInventoryNumber', [None])[0]
				create_host(hostid, desc, mac, ip, inventory_nr)
		else:
			debug(LISTENER, ERROR, '%s: Create not possible, (new == False). new: %r' % (name, new))

	# Object Move
	elif command == 'r':
		debug(LISTENER, INFO, '%s: Detected rename of client, source DN: %s' % (name, dn))
		RENAME = True

# vi: ft=python:tw=79:sw=4:ts=4:noet
